@@ -1,5 +1,5 @@
-# Copyright 2021 AstroLab Software
-# Author: Juliette Vlieghe
+# Copyright 2021-2022 AstroLab Software
+# Author: Juliette Vlieghe, Julien Peloton
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -30,6 +30,223 @@ from astroquery.sdss import SDSS
 
 from fink_science.conversion import dc_mag
 
+def perform_classification(drb, classtar, jd, jdstarthist, ndethist, cdsxmatch, fid,
+        magpsf, sigmapsf, magnr, sigmagnr, magzpsci, isdiffpos, ra, dec, roid):
+    """
+    """
+    high_drb = drb.astype(float) > 0.5
+    high_classtar = classtar.astype(float) > 0.4
+    new_detection = jd.astype(float) - jdstarthist.astype(float) < 0.25
+    not_ztf_sso_candidate = roid.astype(int) != 3
+
+    list_simbad_galaxies = [
+        "galaxy",
+        "Galaxy",
+        "EmG",
+        "Seyfert",
+        "Seyfert_1",
+        "Seyfert_2",
+        "BlueCompG",
+        "StarburstG",
+        "LSB_G",
+        "HII_G",
+        "High_z_G",
+        "GinPair",
+        "GinGroup",
+        "BClG",
+        "GinCl",
+        "PartofG",
+    ]
+
+    keep_cds = \
+        ["Unknown", "Transient", "Fail"] + list_simbad_galaxies
+
+    # Compute DC magnitude
+    mag, err_mag = np.array([
+        dc_mag(i[0], i[1], i[2], i[3], i[4], i[5], i[6])
+        for i in zip(
+            np.array(fid),
+            np.array(magpsf),
+            np.array(sigmapsf),
+            np.array(magnr),
+            np.array(sigmagnr),
+            np.array(magzpsci),
+            np.array(isdiffpos))
+    ]).T
+
+    f_kn = high_drb & high_classtar & new_detection
+    f_kn = f_kn & cdsxmatch.isin(keep_cds) & not_ztf_sso_candidate
+
+    if f_kn.any():
+        # load mangrove catalog
+        curdir = os.path.dirname(os.path.abspath(__file__))
+        mangrove_path = curdir + '/../data/mangrove_filtered.csv'
+        pdf_mangrove = pd.read_csv(mangrove_path)
+
+        catalog_mangrove = SkyCoord(
+            ra=np.array(pdf_mangrove.ra, dtype=float) * u.degree,
+            dec=np.array(pdf_mangrove.dec, dtype=float) * u.degree
+        )
+
+        pdf = pd.DataFrame.from_dict(
+            {
+                'fid': fid[f_kn], 'ra': ra[f_kn],
+                'dec': dec[f_kn], 'mag': mag[f_kn],
+                'err_mag': err_mag[f_kn]
+            }
+        )
+
+        # identify galaxy somehow close to each alert. Distances are in Mpc
+        idx_mangrove, idxself, _, _ = SkyCoord(
+            ra=np.array(pdf.ra, dtype=float) * u.degree,
+            dec=np.array(pdf.dec, dtype=float) * u.degree
+        ).search_around_sky(catalog_mangrove, 2 * u.degree)
+
+        # cross match
+        galaxy_matching = []
+        host_galaxies = []
+        abs_mag_candidate = []
+        host_alert_separation = []
+        for i, row in enumerate(pdf.itertuples()):
+            # SkyCoord didn't keep the original indexes
+            idx_reduced = idx_mangrove[idxself == i]
+            abs_mag = np.array(row.mag - 25 - 5 * np.log10(
+                pdf_mangrove.loc[idx_reduced, :].lum_dist))
+
+            candidates_number = np.argwhere(
+                np.array(
+                    (
+                        SkyCoord(
+                            ra=row.ra * u.degree,
+                            dec=row.dec * u.degree
+                        ).separation(catalog_mangrove[idx_reduced]).radian < 0.01 / pdf_mangrove.loc[idx_reduced, :].ang_dist
+                    ) & (abs_mag > -17) & (abs_mag < -15)
+                )
+            )
+            galaxy_matching.append(len(candidates_number) > 0)
+
+            # save useful information on successful candidates
+            if len(candidates_number) > 0:
+                host_galaxies.append(idx_reduced[candidates_number[0][0]])
+                abs_mag_candidate.append(abs_mag[candidates_number[0][0]])
+                host_alert_separation.append(
+                    SkyCoord(
+                        ra=row.ra * u.degree,
+                        dec=row.dec * u.degree
+                    ).separation(
+                        catalog_mangrove[idx_reduced[candidates_number[0][0]]]
+                    ).radian
+                )
+                # There are sometimes 2 hosts, we currently take the closest
+                # to earth.
+                # This is the index of catalog dataframe and has nothing to do
+                # with galaxies idx.
+
+        f_kn.loc[f_kn] = np.array(galaxy_matching, dtype=bool)
+
+    # check the nature of close objects in SDSS catalog
+    if f_kn.any():
+        no_star = []
+        for i in range(sum(f_kn)):
+            pos = SkyCoord(
+                ra=np.array(ra[f_kn])[i] * u.degree,
+                dec=np.array(dec[f_kn])[i] * u.degree
+            )
+            # for a test on "many" objects, you may wait 1s to stay under the
+            # query limit.
+            table = SDSS.query_region(pos, fields=['type'],
+                                      radius=5 * u.arcsec)
+            type_close_objects = []
+            if table is not None:
+                type_close_objects = table['type']
+            # types: 0: UNKNOWN, 1: STAR, 2: GALAXY, 3: QSO, 4: HIZ_QSO,
+            # 5: SKY, 6: STAR_LATE, 7: GAL_EM
+            to_remove_types = [1, 3, 4, 6]
+            no_star.append(
+                len(np.intersect1d(type_close_objects, to_remove_types)) == 0
+            )
+        f_kn.loc[f_kn] = np.array(no_star, dtype=bool)
+
+    return f_kn, pdf_mangrove, host_galaxies, host_alert_separation, abs_mag_candidate, mag, err_mag
+
+def early_kn_candidates_(
+        drb, classtar, jd, jdstarthist, ndethist, cdsxmatch, fid,
+        magpsf, sigmapsf, magnr, sigmagnr, magzpsci, isdiffpos, ra, dec, roid) -> pd.Series:
+    """ Return alerts considered as KN candidates from the xmatch with Mangrove
+
+    Note the default `data/mangrove_filtered.csv` catalog is loaded.
+
+    Parameters
+    ----------
+    drb: Pandas series
+        Column containing the Deep-Learning Real Bogus score
+    classtar: Pandas series
+        Column containing the sextractor score
+    jd: Pandas series
+        Column containing observation Julian dates at start of exposure [days]
+    jdstarthist: Pandas series
+        Column containing earliest Julian dates corresponding to ndethist
+    ndethist: Pandas series
+        Column containing the number of prior detections (theshold of 3 sigma)
+    cdsxmatch: Pandas series
+        Column containing the cross-match values
+    fid: Pandas series
+        Column containing filter, 1 for green and 2 for red
+    magpsf,sigmapsf: Pandas series
+        Columns containing magnitude from PSF-fit photometry, and 1-sigma error
+    magnr,sigmagnr: Pandas series
+        Columns containing magnitude of nearest source in reference image
+        PSF-catalog within 30 arcsec and 1-sigma error
+    magzpsci: Pandas series
+        Column containing magnitude zero point for photometry estimates
+    isdiffpos: Pandas series
+        Column containing:
+        t or 1 => candidate is from positive (sci minus ref) subtraction;
+        f or 0 => candidate is from negative (ref minus sci) subtraction
+    ra: Pandas series
+        Column containing the right Ascension of candidate; J2000 [deg]
+    dec: Pandas series
+        Column containing the declination of candidate; J2000 [deg]
+    magpsf: Pandas series
+        Column containing the magnitude from PSF-fit photometry [mag]
+    roid: Pandas series
+        Column containing the Solar System label
+
+    Returns
+    -------
+    out: pandas.Series of bool
+        Return a Pandas DataFrame with the appropriate flag:
+        false for bad alert, and true for good alert.
+
+    Examples
+    ----------
+    >>> pdf = pd.read_parquet('datatest')
+    >>> classification = early_kn_candidates_(
+    ...     pdf['candidate'].apply(lambda x: x['drb']),
+    ...     pdf['candidate'].apply(lambda x: x['classtar']),
+    ...     pdf['candidate'].apply(lambda x: x['jd']),
+    ...     pdf['candidate'].apply(lambda x: x['jdstarthist']),
+    ...     pdf['candidate'].apply(lambda x: x['ndethist']),
+    ...     pdf['cdsxmatch'],
+    ...     pdf['candidate'].apply(lambda x: x['fid']),
+    ...     pdf['candidate'].apply(lambda x: x['magpsf']),
+    ...     pdf['candidate'].apply(lambda x: x['sigmapsf']),
+    ...     pdf['candidate'].apply(lambda x: x['magnr']),
+    ...     pdf['candidate'].apply(lambda x: x['sigmagnr']),
+    ...     pdf['candidate'].apply(lambda x: x['magzpsci']),
+    ...     pdf['candidate'].apply(lambda x: x['isdiffpos']),
+    ...     pdf['candidate'].apply(lambda x: x['ra']),
+    ...     pdf['candidate'].apply(lambda x: x['dec']),
+    ...     pdf['roid'])
+    >>> print(pdf[classification]['objectId'].values)
+    []
+    """
+    f_kn, _, _, _, _, _, _ = perform_classification(
+        drb, classtar, jd, jdstarthist, ndethist, cdsxmatch, fid,
+        magpsf, sigmapsf, magnr, sigmagnr, magzpsci, isdiffpos, ra, dec, roid
+    )
+
+    return f_kn
 
 @pandas_udf(BooleanType(), PandasUDFType.SCALAR)
 def early_kn_candidates(
@@ -91,141 +308,16 @@ def early_kn_candidates(
         Return a Pandas DataFrame with the appropriate flag:
         false for bad alert, and true for good alert.
     """
-    high_drb = drb.astype(float) > 0.5
-    high_classtar = classtar.astype(float) > 0.4
-    new_detection = jd.astype(float) - jdstarthist.astype(float) < 0.25
-    not_ztf_sso_candidate = roid.astype(int) != 3
-
     # galactic plane
     gal = SkyCoord(ra.astype(float), dec.astype(float), unit='deg').galactic
 
-    list_simbad_galaxies = [
-        "galaxy",
-        "Galaxy",
-        "EmG",
-        "Seyfert",
-        "Seyfert_1",
-        "Seyfert_2",
-        "BlueCompG",
-        "StarburstG",
-        "LSB_G",
-        "HII_G",
-        "High_z_G",
-        "GinPair",
-        "GinGroup",
-        "BClG",
-        "GinCl",
-        "PartofG",
-    ]
+    out = perform_classification(
+        drb, classtar, jd, jdstarthist, ndethist, cdsxmatch, fid,
+        magpsf, sigmapsf, magnr, sigmagnr, magzpsci, isdiffpos, ra, dec, roid
+    )
 
-    keep_cds = \
-        ["Unknown", "Transient", "Fail"] + list_simbad_galaxies
-
-    # Compute DC magnitude
-    mag, err_mag = np.array([
-        dc_mag(i[0], i[1], i[2], i[3], i[4], i[5], i[6])
-        for i in zip(
-            np.array(fid),
-            np.array(magpsf),
-            np.array(sigmapsf),
-            np.array(magnr),
-            np.array(sigmagnr),
-            np.array(magzpsci),
-            np.array(isdiffpos))
-    ]).T
-
-    f_kn = high_drb & high_classtar & new_detection
-    f_kn = f_kn & cdsxmatch.isin(keep_cds) & not_ztf_sso_candidate
-
-    if f_kn.any():
-        # load mangrove catalog
-        curdir = os.path.dirname(os.path.abspath(__file__))
-        mangrove_path = curdir + '/../data/mangrove_filtered.csv'
-        pdf_mangrove = pd.read_csv(mangrove_path)
-
-        catalog_mangrove = SkyCoord(
-            ra=np.array(pdf_mangrove.ra, dtype=np.float) * u.degree,
-            dec=np.array(pdf_mangrove.dec, dtype=np.float) * u.degree
-        )
-
-        pdf = pd.DataFrame.from_dict(
-            {
-                'fid': fid[f_kn], 'ra': ra[f_kn],
-                'dec': dec[f_kn], 'mag': mag[f_kn],
-                'err_mag': err_mag[f_kn]
-            }
-        )
-
-        # identify galaxy somehow close to each alert. Distances are in Mpc
-        idx_mangrove, idxself, _, _ = SkyCoord(
-            ra=np.array(pdf.ra, dtype=np.float) * u.degree,
-            dec=np.array(pdf.dec, dtype=np.float) * u.degree
-        ).search_around_sky(catalog_mangrove, 2 * u.degree)
-
-        # cross match
-        galaxy_matching = []
-        host_galaxies = []
-        abs_mag_candidate = []
-        host_alert_separation = []
-        for i, row in enumerate(pdf.itertuples()):
-            # SkyCoord didn't keep the original indexes
-            idx_reduced = idx_mangrove[idxself == i]
-            abs_mag = np.array(row.mag - 25 - 5 * np.log10(
-                pdf_mangrove.loc[idx_reduced, :].lum_dist))
-
-            candidates_number = np.argwhere(
-                np.array(
-                    (
-                        SkyCoord(
-                            ra=row.ra * u.degree,
-                            dec=row.dec * u.degree
-                        ).separation(catalog_mangrove[idx_reduced]).radian < 0.01 / pdf_mangrove.loc[idx_reduced, :].ang_dist
-                    ) & (abs_mag > -17) & (abs_mag < -15)
-                )
-            )
-            galaxy_matching.append(len(candidates_number) > 0)
-
-            # save useful information on successful candidates
-            if len(candidates_number) > 0:
-                host_galaxies.append(idx_reduced[candidates_number[0][0]])
-                abs_mag_candidate.append(abs_mag[candidates_number[0][0]])
-                host_alert_separation.append(
-                    SkyCoord(
-                        ra=row.ra * u.degree,
-                        dec=row.dec * u.degree
-                    ).separation(
-                        catalog_mangrove[idx_reduced[candidates_number[0][0]]]
-                    ).radian
-                )
-                # There are sometimes 2 hosts, we currently take the closest
-                # to earth.
-                # This is the index of catalog dataframe and has nothing to do
-                # with galaxies idx.
-
-        f_kn.loc[f_kn] = np.array(galaxy_matching, dtype=bool)
-
-    # check the nature of close objects in SDSS catalog
-    if f_kn.any():
-        no_star = []
-        for i in range(sum(f_kn)):
-            pos = SkyCoord(
-                ra=np.array(ra[f_kn])[i] * u.degree,
-                dec=np.array(dec[f_kn])[i] * u.degree
-                )
-            # for a test on "many" objects, you may wait 1s to stay under the
-            # query limit.
-            table = SDSS.query_region(pos, fields=['type'],
-                                      radius=5 * u.arcsec)
-            type_close_objects = []
-            if table is not None:
-                type_close_objects = table['type']
-            # types: 0: UNKNOWN, 1: STAR, 2: GALAXY, 3: QSO, 4: HIZ_QSO,
-            # 5: SKY, 6: STAR_LATE, 7: GAL_EM
-            to_remove_types = [1, 3, 4, 6]
-            no_star.append(
-                len(np.intersect1d(type_close_objects, to_remove_types)) == 0
-                )
-        f_kn.loc[f_kn] = np.array(no_star, dtype=bool)
+    f_kn, pdf_mangrove, host_galaxies, host_alert_separation, \
+        abs_mag_candidate, mag, err_mag = out
 
     if f_kn.any():
         # Simplify notations
@@ -395,3 +487,15 @@ def early_kn_candidates(
             log.warning(error_message.format('KNWEBHOOK_DWF'))
 
     return f_kn
+
+
+if __name__ == "__main__":
+    """ Execute the test suite """
+    import sys
+    import doctest
+
+    # Numpy introduced non-backward compatible change from v1.14.
+    if np.__version__ >= "1.14.0":
+        np.set_printoptions(legacy="1.13")
+
+    sys.exit(doctest.testmod()[0])
